import os.path
import pandas as pd
import numpy as np
from arch.bootstrap import IIDBootstrap
import plottingMethod as plot
import itertools as it
import math

############# Start Define variables ################
#path to csv file from ARPen
filepath = './Raw/ratings.csv'

#### Reserved Characters
# _,()

#names of the dependent and independent variables
subjectID = 'UserID'
#independent variables with their baseline condition (if there is no baseline condition, use '')
independentV = {'Identifier':''}
#dependent variables with their measurement
dependentV = {'Easy':'Easy','Comfort':'Comfort','Combination':'Combination cube and item selection'}
#name of time based DVs to transform them
timeDV = []
#name of variables that should be added up among trials (e.g. successful selections)
sumDV = []
############# End Define variables ################

############# Start Helper Methods ################
#method to create a dataframe that has the differences between each condition and baseline for each user for a particular independent and dependent Variable
def createMainEffectComparisonDF(start_df, independentVariable, dependentVariable, columnNameAddon = ''):
    #focus on the main effect of the IV by merging potential other IVs together
    if dependentVariable in sumDV:
        start_df = start_df.loc[:, [subjectID, independentVariable, dependentVariable]].groupby([subjectID, independentVariable]).sum().reset_index()
    else:
        start_df = start_df.loc[:, [subjectID, independentVariable, dependentVariable]].groupby([subjectID, independentVariable]).mean().reset_index()
    wide_df = pd.pivot_table(start_df, index=subjectID, columns=independentVariable, values=dependentVariable)
    #get list of different conditions
    conditions = start_df[independentVariable].unique()
    baselineCondition = independentV[independentVariable]
    #if a baseline condition is set: remove baseline condition from conditions list
    if baselineCondition != '':
        conditionList = np.delete(conditions, np.argwhere(conditions==baselineCondition))
    else:
        conditionList = conditions
    #for every condition create a new column with the difference to the baseline condition
    for condition in conditionList:
        if baselineCondition != '':
            newColumnName = condition+'–base_'+dependentVariable
        else:
            newColumnName = condition+'_'+dependentVariable
        if columnNameAddon != '':
            newColumnName = newColumnName + '(' + columnNameAddon + ')'
        if baselineCondition != '':
            wide_df[newColumnName] = wide_df[condition]-wide_df[baselineCondition]
        else:
            wide_df[newColumnName] = wide_df[condition]
    #remove original columns and have only the comparison columns
    wide_df = wide_df.drop(columns=conditions)
    return wide_df

#method to create a dataframe that has the differences between condition and baseline for each user split by the second factor
def createInteractionEffectComparisonDF(start_df, firstFactor, secondFactor, dependentVariable):
    #split dataframe along the values of the secondFactor
    uniqueLevels = start_df[secondFactor].unique()
    comparisonDFs = []
    for level in uniqueLevels:
        filtered_df = start_df[start_df[secondFactor] == level]
        #for each new dataframe, create a main effect comparison df
        comparisonDFs.append(createMainEffectComparisonDF(filtered_df, firstFactor, dependentVariable, columnNameAddon=level))
    
    #merge the resulting dataframes to the interaction df
    interaction_df = comparisonDFs[0].join(comparisonDFs[1:])
    return interaction_df
    
#method to create a confidence interval with bootstrapping using the BCa method to calculate the CI
#return type is a dictionary with mean and lower, upper values for the passed series
def calculateBCaConfInt(x):
    bs = IIDBootstrap(x)
    ci = bs.conf_int(np.mean, 1000, method = 'bca')
    returnValue = {'mean': np.mean(x), 'lower': ci[0][0], 'upper': ci[1][0]}
    return returnValue
############# End Helper Methods ################
        
#load the csv file
original_df = pd.read_csv(filepath)

#Clean up the original csv
#original_df = original_df.drop(columns='CreationTime')

#log transform all timeDVs and add new columns for them
for dv in timeDV:
    newColumnName = 'log('+dv+')'
    original_df[newColumnName] = np.log10(original_df[dv])
    #update dependentV entry to include the log transformed column name
    dependentV[newColumnName] = dependentV.pop(dv)

#Average the individual trials / participant and condition
groupbyList = list(independentV.keys())
groupbyList.insert(0,subjectID)

#create individual lists for the groupby methods
groupbyListMean = groupbyList.copy()
for currentVariable in dependentV:
    if currentVariable not in sumDV:
        groupbyListMean.append(currentVariable)
groupbyListSum = groupbyList.copy() + sumDV

mean_df = original_df.loc[:, groupbyListMean].groupby(groupbyList).mean().reset_index()
sum_df = original_df.loc[:, groupbyListSum].groupby(groupbyList).sum().reset_index()

#merge the two lists to form the main evaluation dataframe (averaged&summed up dependent variables per participant/condition)
combined_df = pd.merge(sum_df, mean_df)

#dictionary to hold all the effect dataframes (main and interaction effects)    
effectDFs = {}
#collect main effect dataframes for every IV and DV combination
for iv in independentV:
    for dv in dependentV:
        effect_df = createMainEffectComparisonDF(combined_df, iv, dv)
        effectDFs[iv+','+dv] = effect_df

#collect interaction effect dataframes for every 2 IVs and DV combination
listOfCombinations = list(it.combinations(independentV, 2))
for combination in listOfCombinations:
    for dv in dependentV:
        interaction_df = createInteractionEffectComparisonDF(combined_df, combination[0], combination[1], dv)
        effectDFs[combination[0]+','+dv+','+combination[1]] = interaction_df
        
#for every main effect dataframe, perform bootstrapping on the series to calculate the confidence intervals        
for effectName in effectDFs:
    effect_df = effectDFs[effectName]
    effectElements = effectName.split(',')
    iv = effectElements[0]
    dv = effectElements[1]
    #check if the current effect_df is of an interaction effect and if so, store the factor in 'by'
    by = ''
    interactionEffect = False
    if len(effectElements) == 3:
        by = effectElements[2]
        interactionEffect = True
    #specify the filename to be used for the data table and the plot
    if interactionEffect:
        filename = iv+','+dv+'_('+by+')'
    else:
        filename = iv+','+dv

    #store calculated dfs in './DataTables' folder
    #check if /DataTables folder exists and create it if not
    if not os.path.exists('./DataTables'):
        os.makedirs('./DataTables')
    fullFilename = './DataTables/(Data)' + filename + '.csv'
    effect_df.to_csv(fullFilename)
    
    #create dictionary with series names (key) and list of confidence interval(s)
    dataDict = {}
    df_columns = list(effect_df)
    for columnHeader in df_columns:
        seriesName = columnHeader.split('_')[0]
        #if this seriesName is not a key of the dataDict yet, create it with an empty list
        if seriesName not in dataDict:
            dataDict[seriesName] = []
        confidenceInterval = calculateBCaConfInt(effect_df[columnHeader])
        if interactionEffect:
            factorLevel = columnHeader.split('(')[1][:-1]
            confidenceInterval['factorLevel'] = factorLevel
            confidenceInterval['factor'] = by
            
        confidenceInterval['seriesName'] = seriesName
        dataDict[seriesName].append(confidenceInterval)
         
    #if current dv has log-transformed values, de-log mean and confidence interval borders
    #anti-log of negative value procedure taken from: https://brainly.in/question/1996764
    currentDataIsLogTransformed = False
    if dv[:4] == 'log(':
        currentDataIsLogTransformed = True
        for key in dataDict:
            listOfCIs = dataDict[key]
            for currentDict in listOfCIs: 
                if currentDict['mean'] < 0:
                    currentDict['mean'] = (10 ** (abs(math.floor(currentDict['mean'])) + currentDict['mean'])) * 10 ** math.floor(currentDict['mean'])
                else:
                    currentDict['mean'] = 10 ** currentDict['mean']
                if currentDict['lower'] < 0:
                    currentDict['lower'] = (10 ** (abs(math.floor(currentDict['lower'])) + currentDict['lower'])) * 10 ** math.floor(currentDict['lower'])
                else:
                    currentDict['lower'] = 10 ** currentDict['lower']        
                if currentDict['upper'] < 0:
                    currentDict['upper'] = (10 ** (abs(math.floor(currentDict['upper'])) + currentDict['upper'])) * 10 ** math.floor(currentDict['upper'])
                else:
                    currentDict['upper'] = 10 ** currentDict['upper']
        newDV = dv[4:-1]
        if dv in dependentV:
            dependentV[newDV] = dependentV.pop(dv)
        dv = newDV
        
    #create plot for current effect
    plot.createPlot(dataDict, filename=filename+'.pdf', xLabel='Difference in '+dependentV[dv], showOnlyIntegers=True, alsoShowPlot = False, ratioData = currentDataIsLogTransformed)

    #store calculated confidence intervals & information in './DataTables' folder
    #combine all confidence intervals into one dataframe
    confIntervalList = []
    for key in dataDict:
        listOfCIs = dataDict[key]
        for currentDict in listOfCIs:
            new_df = pd.DataFrame.from_dict(currentDict, orient='index').transpose()
            new_df.set_index('seriesName', inplace = True)
            confIntervalList.append(new_df)
        
    confInterval_df = pd.concat(confIntervalList)
    #/DataTables folder exists since we have used it before
    fullFilename = './DataTables/(CI)' + filename + '.csv'
    confInterval_df.to_csv(fullFilename)
 